# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""
Trait types for html widgets.
"""

import re
import traitlets
import datetime as dt


_color_names = ['aliceblue', 'antiquewhite', 'aqua', 'aquamarine', 'azure', 'beiae', 'bisque', 'black', 'blanchedalmond', 'blue', 'blueviolet', 'brown', 'burlywood', 'cadetblue', 'chartreuse', 'chocolate', 'coral', 'cornflowerblue', 'cornsilk', 'crimson', 'cyan', 'darkblue', 'darkcyan', 'darkgoldenrod', 'darkgray', 'darkgrey', 'darkgreen', 'darkkhaki', 'darkmagenta', 'darkolivegreen', 'darkorange', 'darkorchid', 'darkred', 'darksalmon', 'darkseagreen', 'darkslateblue', 'darkslategray', 'darkslategrey', 'darkturquoise', 'darkviolet', 'deeppink', 'deepskyblue', 'dimgray', 'dimgrey', 'dodgerblue', 'firebrick', 'floralwhite', 'forestgreen', 'fuchsia', 'gainsboro', 'ghostwhite', 'gold', 'goldenrod', 'gray', 'grey', 'green', 'greenyellow', 'honeydew', 'hotpink', 'indianred ', 'indigo ', 'ivory', 'khaki', 'lavender', 'lavenderblush', 'lawngreen', 'lemonchiffon', 'lightblue', 'lightcoral', 'lightcyan', 'lightgoldenrodyellow', 'lightgray', 'lightgrey', 'lightgreen', 'lightpink', 'lightsalmon', 'lightseagreen', 'lightskyblue', 'lightslategray', 'lightslategrey', 'lightsteelblue', 'lightyellow', 'lime', 'limegreen', 'linen', 'magenta', 'maroon', 'mediumaquamarine', 'mediumblue', 'mediumorchid', 'mediumpurple', 'mediumseagreen', 'mediumslateblue', 'mediumspringgreen', 'mediumturquoise', 'mediumvioletred', 'midnightblue', 'mintcream', 'mistyrose', 'moccasin', 'navajowhite', 'navy', 'oldlace', 'olive', 'olivedrab', 'orange', 'orangered', 'orchid', 'palegoldenrod', 'palegreen', 'paleturquoise', 'palevioletred', 'papayawhip', 'peachpuff', 'peru', 'pink', 'plum', 'powderblue', 'purple', 'rebeccapurple', 'red', 'rosybrown', 'royalblue', 'saddlebrown', 'salmon', 'sandybrown', 'seagreen', 'seashell', 'sienna', 'silver', 'skyblue', 'slateblue', 'slategray', 'slategrey', 'snow', 'springgreen', 'steelblue', 'tan', 'teal', 'thistle', 'tomato', 'transparent', 'turquoise', 'violet', 'wheat', 'white', 'whitesmoke', 'yellow', 'yellowgreen']

# Regex colors #fff and #ffffff
_color_hex = r'#[a-fA-F0-9]{3}(?:[a-fA-F0-9]{3})?'
_color_hex_re = re.compile(fr'^{_color_hex}$')
# Regex colors #ffff and #ffffffff (includes alpha value)
_color_hexa = r'#[a-fA-F0-9]{4}(?:[a-fA-F0-9]{4})?'
_color_hexa_re = re.compile(fr'^{_color_hexa}$')

# Helpers (float percent, int percent with optional surrounding whitespace)
_color_frac_percent = r'\s*(\d+(\.\d*)?|\.\d+)?%?\s*'
_color_int_percent = r'\s*\d+%?\s*'

# rgb(), rgba(), hsl() and hsla() format strings
_color_rgb = r'rgb\({ip},{ip},{ip}\)'
_color_rgba = r'rgba\({ip},{ip},{ip},{fp}\)'
_color_hsl = r'hsl\({fp},{fp},{fp}\)'
_color_hsla = r'hsla\({fp},{fp},{fp},{fp}\)'

# Regex colors rgb/rgba/hsl/hsla
_color_rgbhsl = '({})|({})|({})|({})'.format(
    _color_rgb, _color_rgba, _color_hsl, _color_hsla
).format(ip=_color_int_percent, fp=_color_frac_percent)
_color_rgbhsl_re = re.compile(_color_rgbhsl)

# Support for CSS variables.
# For production rules, see: https://drafts.csswg.org/css-syntax-3/#tokenization

_escape = r'\\([0-9a-fA-F]{1-6}\s?|[^0-9a-fA-F\s])'
_non_ascii = r''.join(
    (
        r'\u00B7',
        r'\u00C0-\u00D6',
        r'\u00C0-\u00D6',
        r'\u00D8-\u00F6',
        r'\u00F8-\u037D',
        r'\u037F-\u1FFF',
        r'\u200C',
        r'\u200D',
        r'\u203F',
        r'\u2040',
        r'\u2070-\u218F',
        r'\u2C00-\u2FEF',
        r'\u3001-\uD7FF',
        r'\uF900-\uFDCF',
        r'\uFDF0-\uFFFD',
        r'\u10000'
    )
)

# Custom CSS identifier
_custom_ident = fr'--([a-zA-Z0-9_\-{_non_ascii}]|{_escape})+'

# Matching for CSS variables with valid color fallback declaration values.
#
# A CSS variable consists of a custom identifier starting with '--'.
# The 'var()' function can be used for substituting the custom property into
# the value of another property.
#
# Here we further restrict the fallback values to be valid colors.

_css_color = fr'({"|".join(_color_names)}|({_color_rgbhsl})|({_color_hex})|({_color_hexa}))'
_css_var_fallback_color = fr'var\({_custom_ident}(,\s*({_css_color}\s*)?)?\)'
_color_var_re = re.compile(_css_var_fallback_color)


class Color(traitlets.Unicode):
    """A string holding a valid HTML color such as 'blue', '#060482', '#A80'"""

    info_text = 'a valid HTML color'
    default_value = traitlets.Undefined

    def validate(self, obj, value):
        if value is None and self.allow_none:
            return value
        if isinstance(value, str):
            if (value.lower() in _color_names or _color_hex_re.match(value) or
                _color_hexa_re.match(value) or _color_rgbhsl_re.match(value) or
                _color_var_re.match(value)):
                return value

        self.error(obj, value)


class Datetime(traitlets.TraitType):
    """A trait type holding a Python datetime object"""

    klass = dt.datetime
    default_value = dt.datetime(1900, 1, 1)


class Date(traitlets.TraitType):
    """A trait type holding a Python date object"""

    klass = dt.date
    default_value = dt.date(1900, 1, 1)


class Time(traitlets.TraitType):
    """A trait type holding a Python time object"""

    klass = dt.date
    default_value = dt.time()


def datetime_to_json(pydt, manager):
    """Serialize a Python datetime object to json.

    Instantiating a JavaScript Date object with a string assumes that the
    string is a UTC string, while instantiating it with constructor arguments
    assumes that it's in local time:

    >>> cdate = new Date('2015-05-12')
    Mon May 11 2015 20:00:00 GMT-0400 (Eastern Daylight Time)
    >>> cdate = new Date(2015, 4, 12) // Months are 0-based indices in JS
    Tue May 12 2015 00:00:00 GMT-0400 (Eastern Daylight Time)

    Attributes of this dictionary are to be passed to the JavaScript Date
    constructor.
    """
    if pydt is None:
        return None
    else:
        try:
            utcdt = pydt.astimezone(dt.timezone.utc)
        except (ValueError, OSError):
            # If year is outside valid range for conversion,
            # use it as-is
            utcdt = pydt
        return dict(
            year=utcdt.year,
            month=utcdt.month - 1,  # Months are 0-based indices in JS
            date=utcdt.day,
            hours=utcdt.hour,  # Hours, Minutes, Seconds and Milliseconds
            minutes=utcdt.minute,  # are plural in JS
            seconds=utcdt.second,
            milliseconds=utcdt.microsecond / 1000,
        )


def datetime_from_json(js, manager):
    """Deserialize a Python datetime object from json."""
    if js is None:
        return None
    else:
        try:
            return dt.datetime(
                js["year"],
                js["month"] + 1,  # Months are 1-based in Python
                js["date"],
                js["hours"],
                js["minutes"],
                js["seconds"],
                js["milliseconds"] * 1000,
            ).astimezone()
        except (ValueError, OSError):
            # If year is outside valid range for conversion,
            # return UTC datetime
            return dt.datetime(
                js["year"],
                js["month"] + 1,  # Months are 1-based in Python
                js["date"],
                js["hours"],
                js["minutes"],
                js["seconds"],
                js["milliseconds"] * 1000,
                dt.timezone.utc,
            )

datetime_serialization = {
    'from_json': datetime_from_json,
    'to_json': datetime_to_json
}


def naive_to_json(pydt, manager):
    """Serialize a naive Python datetime object to json.

    Instantiating a JavaScript Date object with a string assumes that the
    string is a UTC string, while instantiating it with constructor arguments
    assumes that it's in local time:

    >>> cdate = new Date('2015-05-12')
    Mon May 11 2015 20:00:00 GMT-0400 (Eastern Daylight Time)
    >>> cdate = new Date(2015, 4, 12) // Months are 0-based indices in JS
    Tue May 12 2015 00:00:00 GMT-0400 (Eastern Daylight Time)

    Attributes of this dictionary are to be passed to the JavaScript Date
    constructor.
    """
    if pydt is None:
        return None
    else:
        naivedt = pydt.replace(tzinfo=None)
        return dict(
            year=naivedt.year,
            month=naivedt.month - 1,  # Months are 0-based indices in JS
            date=naivedt.day,
            hours=naivedt.hour,  # Hours, Minutes, Seconds and Milliseconds
            minutes=naivedt.minute,  # are plural in JS
            seconds=naivedt.second,
            milliseconds=naivedt.microsecond / 1000,
        )


def naive_from_json(js, manager):
    """Deserialize a naive Python datetime object from json."""
    if js is None:
        return None
    else:
        return dt.datetime(
            js["year"],
            js["month"] + 1,  # Months are 1-based in Python
            js["date"],
            js["hours"],
            js["minutes"],
            js["seconds"],
            js["milliseconds"] * 1000,
        )

naive_serialization = {"from_json": naive_from_json, "to_json": naive_to_json}


def date_to_json(pydate, manager):
    """Serialize a Python date object.

    Attributes of this dictionary are to be passed to the JavaScript Date
    constructor.
    """
    if pydate is None:
        return None
    else:
        return dict(
            year=pydate.year,
            month=pydate.month - 1,  # Months are 0-based indices in JS
            date=pydate.day
        )


def date_from_json(js, manager):
    """Deserialize a Javascript date."""
    if js is None:
        return None
    else:
        return dt.date(
            js['year'],
            js['month'] + 1,  # Months are 1-based in Python
            js['date'],
        )

date_serialization = {
    'from_json': date_from_json,
    'to_json': date_to_json
}

class ByteMemoryView(traitlets.TraitType):
    """A trait for memory views of bytes."""

    default_value = memoryview(b'')
    info_text = 'a memory view object'

    def validate(self, obj, value):
        if isinstance(value, memoryview) and value.format == 'B':
            return value
        self.error(obj, value)

    def default_value_repr(self):
        return repr(self.default_value.tobytes())

class CByteMemoryView(ByteMemoryView):
    """A casting version of the byte memory view trait."""

    def validate(self, obj, value):
        if isinstance(value, memoryview) and value.format == 'B':
            return value

        try:
            mv = memoryview(value)
            if mv.format != 'B':
                mv = mv.cast('B')
            return mv
        except Exception:
            self.error(obj, value)

def time_to_json(pyt, manager):
    """Serialize a Python time object to json."""
    if pyt is None:
        return None
    else:
        return dict(
            hours=pyt.hour,  # Hours, Minutes, Seconds and Milliseconds
            minutes=pyt.minute,  # are plural in JS
            seconds=pyt.second,
            milliseconds=pyt.microsecond / 1000,
        )


def time_from_json(js, manager):
    """Deserialize a Python time object from json."""
    if js is None:
        return None
    else:
        return dt.time(
            js["hours"], js["minutes"], js["seconds"], js["milliseconds"] * 1000
        )


time_serialization = {"from_json": time_from_json, "to_json": time_to_json}


def datetime_to_json(pydt, manager):
    """Serialize a Python datetime object to json.

    Instantiating a JavaScript Date object with a string assumes that the
    string is a UTC string, while instantiating it with constructor arguments
    assumes that it's in local time:

    >>> cdate = new Date('2015-05-12')
    Mon May 11 2015 20:00:00 GMT-0400 (Eastern Daylight Time)
    >>> cdate = new Date(2015, 4, 12) // Months are 0-based indices in JS
    Tue May 12 2015 00:00:00 GMT-0400 (Eastern Daylight Time)

    Attributes of this dictionary are to be passed to the JavaScript Date
    constructor.
    """
    if pydt is None:
        return None
    else:
        try:
            utcdt = pydt.astimezone(dt.timezone.utc)
        except (ValueError, OSError):
            # If year is outside valid range for conversion,
            # use it as-is
            utcdt = pydt
        return dict(
            year=utcdt.year,
            month=utcdt.month - 1,  # Months are 0-based indices in JS
            date=utcdt.day,
            hours=utcdt.hour,  # Hours, Minutes, Seconds and Milliseconds
            minutes=utcdt.minute,  # are plural in JS
            seconds=utcdt.second,
            milliseconds=utcdt.microsecond / 1000,
        )


def datetime_from_json(js, manager):
    """Deserialize a Python datetime object from json."""
    if js is None:
        return None
    else:
        try:
            return dt.datetime(
                js["year"],
                js["month"] + 1,  # Months are 1-based in Python
                js["date"],
                js["hours"],
                js["minutes"],
                js["seconds"],
                js["milliseconds"] * 1000,
                dt.timezone.utc,
            ).astimezone()
        except (ValueError, OSError):
            # If year is outside valid range for conversion,
            # return naive datetime
            return dt.datetime(
                js["year"],
                js["month"] + 1,  # Months are 1-based in Python
                js["date"],
                js["hours"],
                js["minutes"],
                js["seconds"],
                js["milliseconds"] * 1000,
                dt.timezone.utc,
            )


datetime_serialization = {"from_json": datetime_from_json, "to_json": datetime_to_json}



class InstanceDict(traitlets.Instance):
    """An instance trait which coerces a dict to an instance.

    This lets the instance be specified as a dict, which is used
    to initialize the instance.

    Also, we default to a trivial instance, even if args and kwargs
    is not specified."""

    def validate(self, obj, value):
        if isinstance(value, dict):
            return super().validate(obj, self.klass(**value))
        else:
            return super().validate(obj, value)

    def make_dynamic_default(self):
        return self.klass(*(self.default_args or ()),
                          **(self.default_kwargs or {}))


# The regexp is taken
# from https://github.com/d3/d3-format/blob/main/src/formatSpecifier.js
_number_format_re = re.compile(r'^(?:(.)?([<>=^]))?([+\-\( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?([a-z%])?$', re.I)

# The valid types are taken from
# https://github.com/d3/d3-format/blob/main/src/formatTypes.js
_number_format_types = {
    'e', 'f', 'g', 'r', 's', '%', 'p', 'b', 'o', 'd', 'x',
    'X', 'c', ''
}


class NumberFormat(traitlets.Unicode):
    """A string holding a number format specifier, e.g. '.3f'

    This traitlet holds a string that can be passed to the
    `d3-format <https://github.com/d3/d3-format>`_ JavaScript library.
    The format allowed is similar to the Python format specifier (PEP 3101).
    """
    info_text = 'a valid number format'
    default_value = traitlets.Undefined

    def validate(self, obj, value):
        value = super().validate(obj, value)
        re_match = _number_format_re.match(value)
        if re_match is None:
            self.error(obj, value)
        else:
            format_type = re_match.group(9)
            if format_type is None:
                return value
            elif format_type in _number_format_types:
                return value
            else:
                raise traitlets.TraitError(
                    'The type specifier of a NumberFormat trait must '
                    'be one of {}, but a value of \'{}\' was '
                    'specified.'.format(
                        list(_number_format_types), format_type)
                )

class TypedTuple(traitlets.Container):
    """A trait for a tuple of any length with type-checked elements."""
    klass = tuple
    _cast_types = (list,)


def bytes_from_json(js, obj):
    return None if js is None else js.tobytes()

bytes_serialization = {
    'from_json': bytes_from_json,
}



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from pathlib import Path
import sys
import inspect
import warnings

def _get_frame(level):
    """Get the frame at the given stack level."""
    # sys._getframe is much faster than inspect.stack, but isn't guaranteed to
    # exist in all python implementations, so we fall back to inspect.stack()

    # We need to add one to level to account for this get_frame call.
    if hasattr(sys, '_getframe'):
        frame = sys._getframe(level+1)
    else:
        frame = inspect.stack(context=0)[level+1].frame
    return frame


# This function is from https://github.com/python/cpython/issues/67998
# (https://bugs.python.org/file39550/deprecated_module_stacklevel.diff) and
# calculates the appropriate stacklevel for deprecations to target the
# deprecation for the caller, no matter how many internal stack frames we have
# added in the process. For example, with the deprecation warning in the
# __init__ below, the appropriate stacklevel will change depending on how deep
# the inheritance hierarchy is.
def _external_stacklevel(internal):
    """Find the stacklevel of the first frame that doesn't contain any of the given internal strings

    The depth will be 1 at minimum in order to start checking at the caller of
    the function that called this utility method.
    """
    # Get the level of my caller's caller
    level = 2
    frame = _get_frame(level)

    # Normalize the path separators:
    normalized_internal = [str(Path(s)) for s in internal]

    # climb the stack frames while we see internal frames
    while frame and any(s in str(Path(frame.f_code.co_filename)) for s in normalized_internal):
        level +=1
        frame = frame.f_back

    # Return the stack level from the perspective of whoever called us (i.e., one level up)
    return level-1

def deprecation(message, internal='ipywidgets/widgets/'):
    """Generate a deprecation warning targeting the first frame that is not 'internal'
    
    internal is a string or list of strings, which if they appear in filenames in the
    frames, the frames will be considered internal. Changing this can be useful if, for examnple,
    we know that ipywidgets is calling out to traitlets internally.
    """
    if isinstance(internal, str):
        internal = [internal]

    # stack level of the first external frame from here
    stacklevel = _external_stacklevel(internal)

    # The call to .warn adds one frame, so bump the stacklevel up by one
    warnings.warn(message, DeprecationWarning, stacklevel=stacklevel+1)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import mimetypes

from .widget_core import CoreWidget
from .domwidget import DOMWidget
from .valuewidget import ValueWidget
from .widget import register
from traitlets import Unicode, CUnicode, Bool
from .trait_types import CByteMemoryView


@register
class _Media(DOMWidget, ValueWidget, CoreWidget):
    """Base class for Image, Audio and Video widgets.

    The `value` of this widget accepts a byte string.  The byte string is the
    raw data that you want the browser to display.

    If you pass `"url"` to the `"format"` trait, `value` will be interpreted
    as a URL as bytes encoded in UTF-8.
    """

    # Define the custom state properties to sync with the front-end
    value = CByteMemoryView(help="The media data as a memory view of bytes.").tag(sync=True)

    @classmethod
    def _from_file(cls, tag, filename, **kwargs):
        """
        Create an :class:`Media` from a local file.

        Parameters
        ----------
        filename: str
            The location of a file to read into the value from disk.

        **kwargs:
            The keyword arguments for `Media`

        Returns an `Media` with the value set from the filename.
        """
        value = cls._load_file_value(filename)

        if 'format' not in kwargs:
            format = cls._guess_format(tag, filename)
            if format is not None:
                kwargs['format'] = format

        return cls(value=value, **kwargs)

    @classmethod
    def from_url(cls, url, **kwargs):
        """
        Create an :class:`Media` from a URL.

        :code:`Media.from_url(url)` is equivalent to:

        .. code-block: python

            med = Media(value=url, format='url')

        But both unicode and bytes arguments are allowed for ``url``.

        Parameters
        ----------
        url: [str, bytes]
            The location of a URL to load.
        """
        if isinstance(url, str):
            # If str, it needs to be encoded to bytes
            url = url.encode('utf-8')

        return cls(value=url, format='url', **kwargs)

    def set_value_from_file(self, filename):
        """
        Convenience method for reading a file into `value`.

        Parameters
        ----------
        filename: str
            The location of a file to read into value from disk.
        """
        value = self._load_file_value(filename)

        self.value = value

    @classmethod
    def _load_file_value(cls, filename):
        if getattr(filename, 'read', None) is not None:
            return filename.read()
        else:
            with open(filename, 'rb') as f:
                return f.read()

    @classmethod
    def _guess_format(cls, tag, filename):
        # file objects may have a .name parameter
        name = getattr(filename, 'name', None)
        name = name or filename

        try:
            mtype, _ = mimetypes.guess_type(name)
            if not mtype.startswith('{}/'.format(tag)):
                return None

            return mtype[len('{}/'.format(tag)):]
        except Exception:
            return None

    def _get_repr(self, cls):
        # Truncate the value in the repr, since it will
        # typically be very, very large.
        class_name = self.__class__.__name__

        # Return value first like a ValueWidget
        signature = []

        sig_value = 'value={!r}'.format(self.value[:40].tobytes())
        if self.value.nbytes > 40:
            sig_value = sig_value[:-1]+"..."+sig_value[-1]
        signature.append(sig_value)

        for key in super(cls, self)._repr_keys():
            if key == 'value':
                continue
            value = str(getattr(self, key))
            signature.append('{}={!r}'.format(key, value))
        signature = ', '.join(signature)
        return '{}({})'.format(class_name, signature)


@register
class Image(_Media):
    """Displays an image as a widget.

    The `value` of this widget accepts a byte string.  The byte string is the
    raw image data that you want the browser to display.  You can explicitly
    define the format of the byte string using the `format` trait (which
    defaults to "png").

    If you pass `"url"` to the `"format"` trait, `value` will be interpreted
    as a URL as bytes encoded in UTF-8.
    """
    _view_name = Unicode('ImageView').tag(sync=True)
    _model_name = Unicode('ImageModel').tag(sync=True)

    # Define the custom state properties to sync with the front-end
    format = Unicode('png', help="The format of the image.").tag(sync=True)
    width = CUnicode(help="Width of the image in pixels. Use layout.width "
                          "for styling the widget.").tag(sync=True)
    height = CUnicode(help="Height of the image in pixels. Use layout.height "
                           "for styling the widget.").tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_file(cls, filename, **kwargs):
        return cls._from_file('image', filename, **kwargs)

    def __repr__(self):
        return self._get_repr(Image)


@register
class Video(_Media):
    """Displays a video as a widget.

    The `value` of this widget accepts a byte string.  The byte string is the
    raw video data that you want the browser to display.  You can explicitly
    define the format of the byte string using the `format` trait (which
    defaults to "mp4").

    If you pass `"url"` to the `"format"` trait, `value` will be interpreted
    as a URL as bytes encoded in UTF-8.
    """
    _view_name = Unicode('VideoView').tag(sync=True)
    _model_name = Unicode('VideoModel').tag(sync=True)

    # Define the custom state properties to sync with the front-end
    format = Unicode('mp4', help="The format of the video.").tag(sync=True)
    width = CUnicode(help="Width of the video in pixels.").tag(sync=True)
    height = CUnicode(help="Height of the video in pixels.").tag(sync=True)
    autoplay = Bool(True, help="When true, the video starts when it's displayed").tag(sync=True)
    loop = Bool(True, help="When true, the video will start from the beginning after finishing").tag(sync=True)
    controls = Bool(True, help="Specifies that video controls should be displayed (such as a play/pause button etc)").tag(sync=True)

    @classmethod
    def from_file(cls, filename, **kwargs):
        return cls._from_file('video', filename, **kwargs)

    def __repr__(self):
        return self._get_repr(Video)


@register
class Audio(_Media):
    """Displays a audio as a widget.

    The `value` of this widget accepts a byte string.  The byte string is the
    raw audio data that you want the browser to display.  You can explicitly
    define the format of the byte string using the `format` trait (which
    defaults to "mp3").

    If you pass `"url"` to the `"format"` trait, `value` will be interpreted
    as a URL as bytes encoded in UTF-8.
    """
    _view_name = Unicode('AudioView').tag(sync=True)
    _model_name = Unicode('AudioModel').tag(sync=True)

    # Define the custom state properties to sync with the front-end
    format = Unicode('mp3', help="The format of the audio.").tag(sync=True)
    autoplay = Bool(True, help="When true, the audio starts when it's displayed").tag(sync=True)
    loop = Bool(True, help="When true, the audio will start from the beginning after finishing").tag(sync=True)
    controls = Bool(True, help="Specifies that audio controls should be displayed (such as a play/pause button etc)").tag(sync=True)

    @classmethod
    def from_file(cls, filename, **kwargs):
        return cls._from_file('audio', filename, **kwargs)

    def __repr__(self):
        return self._get_repr(Audio)
